/* .vegaplot {
    width: 50%;
} */

/* Make sure action menu isn't cut off */
.vega-embed {
    overflow: visible;
    width: 100% !important;
    height: 100% !important;
    min-height: 0 !important;
}

.vega-embed .fit-x {
    width: 100% !important;
    max-width: 100% !important;
}

/* override altair funny resizing */
.marks {
    width: 100% !important;
    height: 100% !important;

}

/* Remove padding added by has-actions class */
.vega-embed.has-actions {
    padding-right: 0 !important;
}

/* https://discuss.streamlit.io/t/does-altairs-set-embed-options-work-with-streamlit/1675 */
/* remove vega action button*/
details {
    display: none;
}

/* remove solara footer */
.v-application--wrap>div:nth-child(2)>div:nth-child(2) {
    display: none !important;
}

/* Ensure the widget container takes full space */
/* required for molstar expanding in height */
.lm-Widget.p-Widget {
    /* width: 100% !important; */
    height: 100% !important;
    box-sizing: border-box !important;
}


/* Menu styling improvements */
.card-menu-content {
    padding: 8px 16px !important;
    min-width: 200px !important;
}

.card-menu-content .v-input {
    margin-bottom: 8px !important;
}

.card-menu-content .v-input:last-child {
    margin-bottom: 0 !important;
}

/* Reduce space in select fields */
.card-menu-content .v-text-field__details {
    display: none !important;
}

.card-menu-content .v-input--dense .v-input__control {
    min-height: 36px !important;
}

/* make ipypopout strech height */
.card.card--flex>.v-card__text {
    display: flex;
    flex-direction: column;
    min-height: 0;
}

/* html,
body,
#app,
.v-application--wrap,
main.v-content,
.v-content__wrap {
    height: 100% !important;
} */

/* #viewer_container { */
/* height: 100% !important; */
/* min-height: 400px; */
/* fallback minimum if needed */
/* } */


/* Add to your style.css file */
/* .out-of-sync {
    border-left: 3px solid #ff5722 !important;
    background-color: rgba(255, 87, 34, 0.08) !important;
    position: relative;
} */

.out-of-sync::before {
    content: '';
    position: absolute;
    top: 0;
    right: 0;
    width: 6px;
    height: 6px;
    background-color: #ff5722;
    border-radius: 50%;
    margin: 8px;
}